@@ -1,5 +1,4 @@
-/*
- *  linux/tools/build.c
+/* build.c
  *
  *  Copyright (C) 1991, 1992  Linus Torvalds
  */
@@ -22,15 +21,10 @@
  */
 
 #include <stdio.h>	/* fprintf */
-#include <string.h>
-#include <stdlib.h>	/* contains exit */
-#include <sys/types.h>	/* unistd.h needs this */
 #include <sys/stat.h>
-#include <sys/sysmacros.h>
-#include <unistd.h>	/* contains read/write */
 #include <fcntl.h>
 #include <linux/a.out.h>
-#include "mtest.h"
+#include "defs.h"
 
 #define MINIX_HEADER 32
 #define N_MAGIC_OFFSET 1024
