@@ -9,7 +9,7 @@
  */ 
 
 #define SKIP_START	0xa0000
-#define SKIP_END	0xf0000
+#define SKIP_END	0x100000
 #define APM		1		/* 1 = Turn off APM if available */
 #define SCREEN_ADR	0xb8000
 #define SETUPSECS	4		/* Number of setup sectors */
@@ -19,8 +19,8 @@
  * editing the Makefile.
  */
 #define TESTADR		0x1000		/* Final adrs for the test code */
-#define MAINSZ		0x1800		/* Size of primary test code */
-#define TSTSIZE		0x3400		/* Total test size, 2*MAINSZ+0x400 */
+#define MAINSZ		0x3000		/* Size of primary test code */
+#define TSTSIZE		0x6400		/* Total test size, 2*MAINSZ+0x400 */
 #define RELOBASE	0x100000	/* Relocation base address */
 
 #define LOW_ADR		0x00 	        /* Lowest adrs for testing */
